
const { VArray, Variable, Span, Div, P } = require('alkali');
const { onEvent, getElementSize, getElementPosition, styleAttributes } = require('./tools');

class Knob {
	constructor(container, pos, config, callback){

		this.position = pos;
		this.containerLength = config.axisLength;
		this.containerHeight = config.axisHeight;
		this._min = config.min;
		this._max = config.max || this.containerLength;

		this.axisOffset = new Variable(0);
		this.stickOut = new Variable(0); // proportion of knob that sticks out of axis on each side

		this.scaleFactor = Variable.all(config.absoluteMax, config.absoluteMin, this.containerLength).to(
			([max, min, length]) => (max - min) / length
		);

		this.element = new Div('.singleKnob.button', this.getProperties());
		container.appendChild(this.element);

		// bind functions to add/remove event handlers
		this.onMove = this.moveHandler.bind(this);
		this.onDrop = this.disconnect.bind(this);
		this.uiValue = this.position.to(p => ((p + this.stickOut) * this.scaleFactor) + config.absoluteMin);
		if (callback) this.uiValue.subscribe(callback);
		return {
			uiValue: this.uiValue,
			setPosition: this.setPosition.bind(this)
		};
	}

	setPosition(value){ // Public API
		if (!this.attached) return setTimeout(this.setPosition.bind(this, value), 25);
		const scaledValue = this.checkLimits((value / this.scaleFactor) - this.stickOut)
		this.position.put(scaledValue);
	}

	getProperties(){
		const self = this;
		return {
			style: {
				left: this.position.to(p => p + 'px'),
				top: this.axisOffset.to(v => v + 'px')
			},
			attached: function(){self.setDimensions.call(self, this)},
			['on' + onEvent.down]: this.connect.bind(this)
		};
	}

	setDimensions(el){
		const dimentions = getElementSize(el);
		const stickOut = dimentions.width / 2;

		this.min = new Variable(this._min).to(min => min ? min + stickOut : 0 - stickOut);
		this.max = this._max.to(max => max - stickOut);

		this.stickOut.put(stickOut);
		this.position.put(this.min);
		this.axisOffset.put((dimentions.height - this.containerHeight) / 2 * -1);
		this.attached = true;
	}

	checkLimits(val){
		if (val < this.min) val = this.min.valueOf();
		else if (val > this.max) val = this.max.valueOf();
		return val;
	}

	connect(e){
		if (!this.element) return; // if we add the event handler after attach this should net be needed
		e.preventDefault();
		const touchDown = this.getPointerCoords(e);
		const knobPosition = this.element.getBoundingClientRect();
		const parentCoords = getElementPosition(this.element);

		this.parentCoords = {
			x: parentCoords.x + this.stickOut,
			y: parentCoords.y
		}
		this.pointerOffset = {
			x: knobPosition.left - touchDown.x,
			y: knobPosition.top - touchDown.y
		}

		window.addEventListener(onEvent.move, this.onMove);
		window.addEventListener(onEvent.up, this.onDrop);
	}

	disconnect(){
		window.removeEventListener(onEvent.move, this.onMove);
		window.removeEventListener(onEvent.up, this.onDrop);
		this.parentCoords = this.lastCoords = null;
	}

	getPointerCoords(e) {
		let coords = {
			x: e.pageX || (e.touches && e.touches[0].pageX) || this.lastCoords.x,
			y: e.pageY || (e.touches && e.touches[0].pageY) || this.lastCoords.y
		};
		// to fix buggy behavior when mouse leaves window
		this.lastCoords = coords;
		return coords;
	}

	moveHandler(e){
		e.preventDefault();
        if (!this.parentCoords) return;
		const coords = this.getPointerCoords(e);
		const value = this.checkLimits(coords.x - this.parentCoords.x);
        this.position.put(value);
	}
}

module.exports = Knob;
